import { combineReducers } from 'redux';
import conformsTo from 'lodash/conformsTo';
import isFunction from 'lodash/isFunction';
import isObject from 'lodash/isObject';
import invariant from 'invariant';
import isEmpty from 'lodash/isEmpty';
import isString from 'lodash/isString';
import React, { Component } from 'react';
import PropTypes from 'prop-types';
import hoistNonReactStatics from 'hoist-non-react-statics';

function _extends() {
  _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  return _extends.apply(this, arguments);
}

function _inheritsLoose(subClass, superClass) {
  subClass.prototype = Object.create(superClass.prototype);
  subClass.prototype.constructor = subClass;
  subClass.__proto__ = superClass;
}

var REDUCER_CREATOR_SYMBOL = Symbol('createReducer');

/**
 * A store enhancer that enables dynamic injection of reducers.
 *
 * @param {function} createReducer function for creating root reducer.
 * @returns {function} A store enhancer applying the middleware.
 */

function applyAsyncReducers(createReducer) {
  return function (createStore) {
    return function (reducer, preloadedState) {
      var _extends2;

      var store = createStore(reducer, preloadedState);
      return _extends({}, store, (_extends2 = {
        injectedReducers: {}
      }, _extends2[REDUCER_CREATOR_SYMBOL] = createReducer, _extends2));
    };
  };
}

/**
 * A helper for creating the root reducer factory.
 *
 * @param reducers
 * @return {function(*): Reducer<any>}
 */

function composeReducerCreator(reducers) {
  return function (asyncReducers) {
    return combineReducers(_extends({}, reducers, asyncReducers));
  };
}

/**
 * Validate the shape of redux store for dynamic reducer injection.
 */

function checkStoreWithAsyncReducers(store) {
  invariant(conformsTo(store, {
    dispatch: isFunction,
    subscribe: isFunction,
    getState: isFunction,
    replaceReducer: isFunction,
    injectedReducers: isObject
  }) && isFunction(store[REDUCER_CREATOR_SYMBOL]), '(react-observatory) injectReducer: Expected a valid redux store');
}

/**
 * Internal helper function for extracting the injector for epics.
 *
 * @param store
 * @param isValid
 * @return {injectReducer}
 */

function injectReducerFactory(store, isValid) {
  return function (key, reducer) {
    if (!isValid) checkStoreWithAsyncReducers(store);
    invariant(isString(key) && !isEmpty(key) && isFunction(reducer), '(react-observatory) injectReducer: Expected `reducer` to be a reducer function'); // Check `store.injectedReducers[key] === reducer` for hot reloading when a key is the same but a reducer is different

    if (Reflect.has(store.injectedReducers, key) && store.injectedReducers[key] === reducer) {
      return;
    }

    var reducerCreator = store[REDUCER_CREATOR_SYMBOL];
    store.injectedReducers[key] = reducer;
    store.replaceReducer(reducerCreator(store.injectedReducers));
  };
}
function getInjectors(store) {
  checkStoreWithAsyncReducers(store);
  return {
    injectReducer: injectReducerFactory(store, true)
  };
}

/**
 * High-Order Component that dynamically injects a reducer.
 *
 * @param {string} key A key of the reducer
 * @param {function} reducer A reducer that will be injected
 *
 */

/**
 * High-Order Component that dynamically injects a reducer.
 *
 * @param key {string} A key of the reducer
 * @param reducer {function} reducer A reducer that will be injected
 * @return {function(*)} HOC with a reducer.
 */

var injectReducer = (function (key, reducer) {
  return function (WrappedComponent) {
    var ReducerInjector =
    /*#__PURE__*/
    function (_Component) {
      _inheritsLoose(ReducerInjector, _Component);

      function ReducerInjector() {
        var _this;

        for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
          args[_key] = arguments[_key];
        }

        _this = _Component.call.apply(_Component, [this].concat(args)) || this;
        _this.injectors = getInjectors(_this.context.store);
        return _this;
      }

      var _proto = ReducerInjector.prototype;

      _proto.componentWillMount = function componentWillMount() {
        var injectReducer = this.injectors.injectReducer;
        injectReducer(key, reducer);
      };

      _proto.render = function render() {
        return React.createElement(WrappedComponent, this.props);
      };

      return ReducerInjector;
    }(Component);

    ReducerInjector.WrappedComponent = WrappedComponent;
    ReducerInjector.contextTypes = {
      store: PropTypes.object.isRequired
    };
    ReducerInjector.displayName = "withReducer(" + (WrappedComponent.displayName || WrappedComponent.name || 'Component') + ")";
    return hoistNonReactStatics(ReducerInjector, WrappedComponent);
  };
});

// Store Enhancers.

export { applyAsyncReducers, composeReducerCreator, injectReducer };
